
#include "TFMatrix.h"





//-----------------------------------------------------------------//
// 3次元ベクトルの格納クラス
// 本来はＮ次元ベクトルの格納クラスを目標として作ったが、現状、３次元までしか対応していないが、
// ほとんどのインターフェース関数は、Ｎ次元に対応出来るように設計されている。
//-----------------------------------------------------------------//



//--------------------------------------------------//
// boostのマトリックスに変換する
//--------------------------------------------------//
boost::multi_array<std::complex<double> , 3>*  TFMatrix::createBoostMatrix()
{
    try
    {
        std::vector<std::pair<int,int> > range_list;
        TFMatrix::all_range();
        range_list = TFMatrix::all_range();

        delete boost_matrix;

        int range1 = range_list[0].second - range_list[0].first + 1;
        int range2 = range_list[1].second - range_list[1].first + 1;
        int range3 = range_list[2].second- range_list[2].first + 1;

        boost::multi_array<std::complex<double>, 3> *boost_temp
                   = new boost::multi_array<std::complex<double>, 3> (boost::extents[range1][range2][range3]);

        for (int i = range_list[0].first; i <= range_list[0].second; i++) {
            for (int j = range_list[1].first; j <= range_list[1].second; j++) {
                for (int k = range_list[2].first; k <= range_list[2].second; k++) {
                    (*boost_temp)[i][j][k] = matrix[i][j][k];
                }
            }
        }

        return this->boost_matrix = boost_temp;

    }
    catch (const std::exception& e)
    {
         THROW_CRITICAL(e.what());
    }
    catch (...)
    {
        THROW_CRITICAL(("unexpected error"));
    }


}







    //--------------------------------------------------//
    //! 各次元とその深さを表示する
    //--------------------------------------------------//
    void TFMatrix::print_size()
    {
        for(unsigned int i = 0; i < dimension.size(); i++)
        {
            std::cout << "dim=" << i << " depth=" << dimension[i] << std::endl;
        }
        std::cout << std::endl;
    }

    //--------------------------------------------------//
    //! 各次元毎に、計算範囲(range)を表示する
    //--------------------------------------------------//
    void TFMatrix::print_range(std::vector<std::pair<int,int> >range_list)
    {
        for(unsigned int i = 0; i < range_list.size(); i++) {
            std::cout << "dim=" << i << " range=" << range_list[i].first << "-" << range_list[i].second << std::endl;
        }
    }

    //--------------------------------------------------//
    //! 全てのマトリックス要素を表示する
    //--------------------------------------------------//
    void TFMatrix::print()
    {
        std::vector<std::pair<int,int> >range_list;
        range_list = all_range();
        print(range_list);
    }


    //--------------------------------------------------//
    //! rangeで指定された範囲のマトリックス要素を表示する
    //--------------------------------------------------//
    void TFMatrix::print(std::vector<std::pair<int,int> >range_list)
    {
        for(int i = range_list[0].first; i <= range_list[0].second; i++) {
            for(int j = range_list[1].first; j <= range_list[1].second; j++) {
                for(int k = range_list[2].first; k <= range_list[2].second; k++) {
                    std::cout << i << "," << j << "," << k << "=" << matrix[i][j][k] << std::endl;
                }
            }
        }
    }





    //----------------------------------------------------------------------//
    //! \name マトリックスのサイズを変える関数一覧
    /*!
     * マトリックスのサイズが強制的にリサイズされるだけで内容に関しては保障されないことに注意する
     * 事実上、クリア状態からのリサイズに使用する。
     */
    //\{
    //----------------------------------------------------------------------//

    //--------------------------------------------------//
    //! 可変引数で変更するサイズを与える
    //--------------------------------------------------//
    void TFMatrix::resize(int argc, ...)
    {
        try
        {
            std::vector<int> size;
            // 引数処理
            va_list ap;
            va_start(ap, argc);
            vresize(argc, ap);
            va_end(ap);

        }
        catch (const std::exception& e)
        {
             THROW_CRITICAL(e.what());
        }
        catch (...)
        {
            THROW_CRITICAL(("unexpected error"));
        }
    }


    //--------------------------------------------------//
    //! マトリックスのサイズ自体を０にする
    //--------------------------------------------------//
    void TFMatrix::clear()
    {
        matrix.clear();
        operator_range.clear();
        dimension.clear();
    }



    //--------------------------------------------------//
    //! resize のva_listを使う内部関数
    //--------------------------------------------------//
    void TFMatrix::vresize(int argc, va_list ap)
    {
        std::vector<int> size_list;
        // 引数処理
        size_list = varg_vector(argc, ap);
        resize(size_list);
    }


    //--------------------------------------------------//
    //! resize の可変引数を使用しない内部関数
    /*!
     * rangeで指定された範囲と同じ大きさになるだけで、
     * その範囲の要素の内容は保持されない。
     */
    //--------------------------------------------------//
    void TFMatrix::resize_by_range(std::vector<std::pair<int, int> > range)
    {
        std::vector<int> list;

        list.resize(range.size());
        for (unsigned int i=0; i<range.size(); i++) {
            list[i] = range[i].second - range[i].first + 1;
        }
        resize(list);
    }


    //--------------------------------------------------//
    //! 最終的にこの関数が呼ばれてリサイズが行われる
    //--------------------------------------------------//
    void TFMatrix::resize(std::vector<int> size)
    {
        try
        {
            matrix.resize(size[0]);
            for(int i = 0; i < size[0]; i++) {
                matrix[i].resize(size[1]);
                for(int j = 0; j < size[1]; j++) {
                    matrix[i][j].resize(size[2]);
                }
            }

            dimension = size; // 注意

        }
        catch (const std::exception& e)
        {
             THROW_CRITICAL(e.what());
        }
        catch (...)
        {
            THROW_CRITICAL(("unexpected error"));
        }
    }



    //--------------------------------------------------//
    //! operator_rangeをクリアする関数(内部的に使う関数)
    //--------------------------------------------------//
    void TFMatrix::clear_range()
    {
        operator_range.clear();
    }


    //--------------------------------------------------//
    //! 単に各次元の深さが入った配列を返す
    //--------------------------------------------------//
    std::vector<int> TFMatrix::size()
    {
        return dimension;
    }

    //--------------------------------------------------//
    //! \name 切り出し関数
    //\{
    //!マトリックスの指定された領域を切りだして取得。
    /*! operator_range関数で領域を指定するのではない事に注意
     * 右辺値にはなれない事に注意する、領域をコピーするには左辺値のマトリックスの operator_rangeを空にして代入演算を行う
     */
    //--------------------------------------------------//
    void TFMatrix::trim(int argc, ...)
    {
        try
        {
            va_list ap;
            va_start(ap, argc);
            trim(varg_range(argc, ap));
            va_end(ap);
            return;

        }
        catch (const std::exception& e)
        {
             THROW_CRITICAL(e.what());
        }
        catch (...)
        {
            THROW_CRITICAL(("unexpected error"));
        }
    }


    //--------------------------------------------------//
    //!マトリックスの指定された次元の指定範囲を切り捨てる
    /*!ある次元の全範囲を削除すると、全次元を削除した事になってしまうので削除できない
     * rangeが全範囲指定の場合、その次元に対して、削除を行わない事を意味する
     */
    //--------------------------------------------------//
    void TFMatrix::cut_trim(int argc, ...)
    {
        try
        {
            va_list ap;
            va_start(ap, argc);
            cut_trim(varg_range(argc, ap));
            va_end(ap);
            return;

        }
        catch (const std::exception& e)
        {
             THROW_CRITICAL(e.what());
        }
        catch (...)
        {
            THROW_CRITICAL(("unexpected error"));
        }

    }


    //--------------------------------------------------//
    //--------------------------------------------------//
    void TFMatrix::trim(std::vector<std::pair<int,int> >range_list)
    {
        try
        {
            // リサイズ元をテンポラリーにコピー
            TFMatrix tmp_matrix;
            tmp_matrix = (*this);

            // リサイズ元をリサイズ
            this->resize_by_range(range_list);

            // コピーからリサイズ元に戻す
            for(int i = range_list[0].first; i <= range_list[0].second; i++)
            {
                for(int j = range_list[1].first; j <= range_list[1].second; j++)
                {
                    for(int k = range_list[2].first; k <= range_list[2].second; k++)
                    {
                        matrix[i-range_list[0].first][j-range_list[1].first][k-range_list[2].first] = tmp_matrix.matrix[i][j][k];
                    }
                }
            }

        }
        catch (const std::exception& e)
        {
             THROW_CRITICAL(e.what());
        }
        catch (...)
        {
            THROW_CRITICAL(("unexpected error"));
        }

    }


    //--------------------------------------------------//
    //
    //--------------------------------------------------//
    void TFMatrix::cut_trim(std::vector<std::pair<int,int> >range_list)
    {
        try
        {
            for(unsigned int dim=0; dim < range_list.size(); dim++)
            {
                int from = range_list[dim].first;
                int to   = range_list[dim].second;

                switch(dim)
                {
                case(0):
                {
                    if (from == 0 &&   to + 1 >= dimension[dim])
                    {
                        continue; // 各次元の全てが指定された場合は実行を行わない
                    }

                    typename std::vector<std::vector<std::vector<std::complex<double> > > >::iterator it = matrix.begin();
                    it+=from;
                    for(int cnt=0; cnt < to - from + 1; cnt++) {
                        it = matrix.erase(it);
                    }
                    dimension[0]-=(to-from+1); // サイズの更新
                    break;
                }
                case(1):
                {
                    if (from == 0 &&   to + 1 >= dimension[dim]) {
                        continue; // 各次元の全てが指定された場合は実行を行わない
                    }
                    for(int i = 0; i < dimension[0]; i++) {
                        typename std::vector<std::vector<std::complex<double> > >::iterator it = matrix[i].begin();
                        it+=from;
                        for(int cnt=0; cnt < to - from + 1; cnt++) {
                            it = matrix[i].erase(it);
                        }
                    }
                    dimension[1]-=(to-from+1); // サイズの更新
                    break;
                }
                case(2):
                {
                    if (from == 0 &&   to + 1 >= dimension[dim])
                    {
                        continue; // 各次元の全てが指定された場合は実行を行わない
                    }
                    for(int i = 0; i < dimension[0]; i++) {
                        for(int j = 0; j < dimension[1]; j++) {
                            typename  std::vector<std::complex<double> >::iterator it = matrix[i][j].begin();
                            it+=from;
                            for(int cnt=0; cnt < to - from + 1; cnt++) {
                                it = matrix[i][j].erase(it);
                            }
                        }
                    }
                    dimension[2]-=(to-from+1); // サイズの更新
                    break;
                }//case
                }//switch
            }//for

        }
        catch (const std::exception& e)
        {
             THROW_CRITICAL(e.what());
        }
        catch (...)
        {
            THROW_CRITICAL(("unexpected error"));
        }


    }




    //--------------------------------------------------//
    //! operatorの演算対象の範囲をセットする。
    //--------------------------------------------------//
    TFMatrix&  TFMatrix::range(int argc, ...)
    {
        va_list ap;
        va_start(ap, argc);
        operator_range = varg_range(argc, ap);
        va_end(ap);
        return (*this);
    }

    //--------------------------------------------------//
    //! 引数リストからrangeを生成する
    //--------------------------------------------------//
    std::vector<std::pair<int,int> > TFMatrix::varg_range(int argc, va_list ap)
    {
        try
        {
            // 戻り値
            std::vector<std::pair<int,int> > result;

            // 全範囲を格納
            result = all_range();

            // 引数不一致
            if (argc%3) { // A,from,to A次元目がfromからtoまで
                return result;
            }

            int dim;
            for(int i = 0; i < argc; i+=3)
            {
                dim = va_arg(ap, int);
                result[dim].first  = va_arg(ap, int);
                result[dim].second = va_arg(ap, int);
            }
            return result;


        }
        catch (const std::exception& e)
        {
             THROW_CRITICAL(e.what());
        }
        catch (...)
        {
            THROW_CRITICAL(("unexpected error"));
        }

    }


/*
    //--------------------------------------------------//
    //! 引数リストから、ベクターを生成する
    //--------------------------------------------------//
    std::vector<int> TFMatrix::arg_vector(int argc, ...)
    {
        std::vector<int> result;

        va_list ap;
        va_start(ap, argc);
        result = varg_vector(argc, ap);
        va_end(ap);

        return ap;
    }
*/

    //--------------------------------------------------//
    //! 引数リストから、ベクターを生成する
    //--------------------------------------------------//
    std::vector<int> TFMatrix::varg_vector(int argc, va_list ap)
    {
        try
        {
            // 戻り値
            std::vector<int> result;
            result.resize(argc);

            for(int i = 0; i < argc; i++) {
                result[i] = va_arg(ap, int);
            }
            return result;

        }
        catch (const std::exception& e)
        {
             THROW_CRITICAL(e.what());
        }
        catch (...)
        {
            THROW_CRITICAL(("unexpected error"));
        }

    }


    //--------------------------------------------------//
    //--------------------------------------------------//
    TFMatrix& TFMatrix::range(std::vector<std::pair<int,int> >range)
    {
        operator_range = range;
        return (*this);
    }


    //--------------------------------------------------//
    //! operator_range自体の取得
    //--------------------------------------------------//
    std::vector<std::pair<int,int> > TFMatrix::get_range()
    {
        return operator_range;
    }

    //--------------------------------------------------//
    //! 全範囲のrange値を取得
    //--------------------------------------------------//
    std::vector<std::pair<int,int> > TFMatrix::all_range()
    {
        std::vector<std::pair<int,int> > result;
            result.resize(dimension.size());

            // 初期値として全領域を与える
            for (unsigned int i = 0; i < dimension.size(); i++) {
                result[i].first  = 0;
                result[i].second = dimension[i] -1;
            }
            return result;
    }

    //--------------------------------------------------//
    // 引数リストからrangeを生成する
    // 例 arg_range(6,1,0,3,2,1,1);
    //  引数6個 1次元目は0〜3 2次元目は1〜1 それ以外の次元は全範囲
    //--------------------------------------------------//
    std::vector<std::pair<int,int> > TFMatrix::arg_range(int argc, ...)
    {
        try
        {
            std::vector<std::pair<int,int> > result;

            va_list ap;
            va_start(ap, argc);
            result = varg_range(argc, ap);
            va_end(ap);

            return result;

        }
        catch (const std::exception& e)
        {
             THROW_CRITICAL(e.what());
        }
        catch (...)
        {
            THROW_CRITICAL(("unexpected error"));
        }

    }







    //--------------------------------------------------//
    //! 位相 Complex専用関数である事に注意する
    //--------------------------------------------------//
    TFMatrix& TFMatrix::arg()
    {
        try
        {
            if(get_range().size() == 0)
            {
                operator_range = all_range();
            }
            std::vector<std::pair<int,int> > &range_list = operator_range;
            for(int i = range_list[0].first; i <= range_list[0].second; i++)
            {
                for(int j = range_list[1].first; j <= range_list[1].second; j++)
                {
                    for(int k = range_list[2].first; k <= range_list[2].second; k++)
                    {
                            matrix[i][j][k] = std::arg(matrix[i][j][k]);
                    }
                }
            }
            clear_range();
            return *this;

        }
        catch (const std::exception& e)
        {
             THROW_CRITICAL(e.what());
        }
        catch (...)
        {
            THROW_CRITICAL(("unexpected error"));
        }
    }



    //--------------------------------------------------//
    //! dB(Log10*abs*10) この関数はComplex専用の関数である事に注意
    //--------------------------------------------------//
    TFMatrix& TFMatrix::db()
    {
        try
        {
            if(get_range().size() == 0)
            {
                operator_range = all_range();
            }

            std::vector<std::pair<int,int> > &range_list = operator_range;
            for(int i = range_list[0].first; i <= range_list[0].second; i++)
            {
                for(int j = range_list[1].first; j <= range_list[1].second; j++)
                {
                    for(int k = range_list[2].first; k <= range_list[2].second; k++)
                    {
                        std::complex<double> &temp = matrix[i][j][k];
                        // abs,real()などcomplexの関数が呼び出されている
                        temp = std::abs(temp);
                        temp = std::log10(temp.real()) * 10;
                        matrix[i][j][k] = temp;
                    }
                }
            }
            clear_range();
            return *this;

        }
        catch (const std::exception& e)
        {
             THROW_CRITICAL(e.what());
        }
        catch (...)
        {
            THROW_CRITICAL(("unexpected error"));
        }
    }



    //--------------------------------------------------//
    //! 合計
    //--------------------------------------------------//
    std::complex<double> TFMatrix::sum()
    {
        try
        {
            std::vector<std::pair<int,int> >temp;
            if(get_range().size() == 0)
            {
                temp = all_range();
            }
            else
            {
                temp = operator_range;
            }

            clear_range();
            std::vector<std::pair<int,int> > &range_list = temp;

            std::complex<double> total = 0;
            for(int i = range_list[0].first; i <= range_list[0].second; i++)
            {
                for(int j = range_list[1].first; j <= range_list[1].second; j++)
                {
                    for(int k = range_list[2].first; k <= range_list[2].second; k++)
                    {
                        total+=(matrix[i][j][k]);
                    }
                }
            }
            return total;

        }
        catch (const std::exception& e)
        {
             THROW_CRITICAL(e.what());
        }
        catch (...)
        {
            THROW_CRITICAL(("unexpected error"));
        }

    }


    //--------------------------------------------------//
    //! 相加平均
    //--------------------------------------------------//
    std::complex<double>  TFMatrix::mean()
    {
        try
        {
            std::vector<std::pair<int,int> >temp;
            if(get_range().size() == 0)
            {
                temp = all_range();
            }
            else
            {
                temp = operator_range;
            }
            clear_range();
            std::vector<std::pair<int,int> > &range_list = temp;

            std::complex<double> total = 0;
            int sum_count =0;

            for(int i = range_list[0].first; i <= range_list[0].second; i++)
            {
                for(int j = range_list[1].first; j <= range_list[1].second; j++)
                {
                    for(int k = range_list[2].first; k <= range_list[2].second; k++)
                    {
                        total+=(matrix[i][j][k]);
                        sum_count++;
                    }
                }
            }
            return total/(std::complex<double>)sum_count;

        }
        catch (const std::exception& e)
        {
             THROW_CRITICAL(e.what());
             return 0;
        }
        catch (...)
        {
            THROW_CRITICAL(("unexpected error"));
            return 0;
        }

    }




    //--------------------------------------------------//
    //! 絶対値 複素数でもＯＫ
    //--------------------------------------------------//
    TFMatrix& TFMatrix::abs()
    {
        try
        {
            if(get_range().size() == 0)
            {
                operator_range = all_range();
            }

            std::vector<std::pair<int,int> > &range_list = operator_range;

            for(int i = range_list[0].first; i <= range_list[0].second; i++)
            {
                for(int j = range_list[1].first; j <= range_list[1].second; j++)
                {
                    for(int k = range_list[2].first; k <= range_list[2].second; k++)
                    {
                        matrix[i][j][k] = std::abs(matrix[i][j][k]);
                    }
                }
            }
            return *this;

        }
        catch (const std::exception& e)
        {
             THROW_CRITICAL(e.what());
        }
        catch (...)
        {
            THROW_CRITICAL(("unexpected error"));
        }

    }


    //--------------------------------------------------//
    // 最大値
    // * operator_range範囲の最大値を返す
    //  * offsetには、配列の添え字が戻るが、２次元以上になった時のoffsetの数え方は、
    // MATLABと異なるので注意する
    // MAXとなる同値の数字があった場合は、最も小さいoffsetが戻る
    // 戻り値には、対象となる要素の値が戻る
    //
    //--------------------------------------------------//
     std::complex<double>  TFMatrix::complex_max(int *offset)
    {
        try
         {
            if(get_range().size() == 0)
            {
                operator_range = all_range();
            }

            std::complex<double> max_temp = matrix[0][0][0]; // point
            int max_offset = 0;
            int count=0;
            std::vector<std::pair<int,int> > &range_list = operator_range;

            for(int i = range_list[0].first; i <= range_list[0].second; i++)
            {
                for(int j = range_list[1].first; j <= range_list[1].second; j++)
                {
                    for(int k = range_list[2].first; k <= range_list[2].second; k++)
                    {
                        if (matrix[i][j][k].real() > max_temp.real())
                        {
                            max_temp = matrix[i][j][k];
                            max_offset=count;
                        }
                        count++;
                    }
                }
            }

            if(offset)
            {
                *offset = max_offset;
            }
            return max_temp;


         }
         catch (const std::exception& e)
         {
              THROW_CRITICAL(e.what());
         }
         catch (...)
         {
             THROW_CRITICAL(("unexpected error"));
         }
     }



    //--------------------------------------------------//
    //! 複素共役 (単項演算子)
    //--------------------------------------------------//
    TFMatrix& TFMatrix::conj()
    {
        try
        {
            if(get_range().size() == 0)
            {
                operator_range = all_range();
            }

            std::vector<std::pair<int,int> > &range_list = operator_range;
            for(int i = range_list[0].first; i <= range_list[0].second; i++)
            {
                for(int j = range_list[1].first; j <= range_list[1].second; j++)
                {
                    for(int k = range_list[2].first; k <= range_list[2].second; k++)
                    {
                        matrix[i][j][k] = std::conj(matrix[i][j][k]);
                    }
                }
            }
            clear_range();

            return *this;

        }
        catch (const std::exception& e)
        {
             THROW_CRITICAL(e.what());
        }
        catch (...)
        {
            THROW_CRITICAL(("unexpected error"));
        }


    }


    //--------------------------------------------------//
    //! べき乗(単項演算子)
    //--------------------------------------------------//
    TFMatrix& TFMatrix::pow(int index)
    {
        try
        {
            if(get_range().size() == 0)
            {
                operator_range = all_range();
            }
            std::vector<std::pair<int,int> > &range_list = operator_range;
            for(int i = range_list[0].first; i <= range_list[0].second; i++)
            {
                for(int j = range_list[1].first; j <= range_list[1].second; j++)
                {
                    for(int k = range_list[2].first; k <= range_list[2].second; k++)
                    {
                        matrix[i][j][k] = std::pow(matrix[i][j][k], index);
                    }
                }
            }
            clear_range();
            return *this;

        }
        catch (const std::exception& e)
        {
             THROW_CRITICAL(e.what());
        }
        catch (...)
        {
            THROW_CRITICAL(("unexpected error"));
        }

    }





    //---------------------------------------------------//
    //　エラーダイアログ
    //---------------------------------------------------//
    void TFMatrix::THROW_CRITICAL(QString msg)
    {
        QMessageBox  msgBox;
        msgBox.setIcon(QMessageBox::Critical);
        msgBox.setText(msg);
        msgBox.exec();
    }




/*
    //--------------------------------------------------//
    //! 逆正接(単項演算子)
    //--------------------------------------------------//
    TFMatrix&  TFMatrix::atan2(TFMatrix &rhs)
    {

         TFMatrix &lhs = (*this);
        init_operator_range(&lhs, &rhs);
 //       IFDEBUG(check_dimension(&lhs, &rhs));

        // 暫定処置
        int i_start =lhs.operator_range[0].first;
        int j_start =lhs.operator_range[1].first;
        int k_start =lhs.operator_range[2].first;
        int rhs_i_start = rhs.operator_range[0].first;
        int rhs_j_start = rhs.operator_range[1].first;
        int rhs_k_start = rhs.operator_range[2].first;

        for(int i = operator_range[0].first; i <= operator_range[0].second; i++)
        {
            for(int j = operator_range[1].first; j <= operator_range[1].second; j++)
            {
                for(int k = operator_range[2].first; k <= operator_range[2].second; k++)
                {

//                    lhs.matrix[i][j][k]
//                            = atan2(lhs.matrix[i][j][k], rhs.matrix[i-i_start + rhs_i_start][j-j_start + rhs_j_start][k-k_start + rhs_k_start]);
                }
            }
        }
        finish_operator_range(&lhs, &rhs);
        return *this;
    }

*/






